"""时间序列计算层、神经网络模型定义.

复现华泰金工 alpha net V2、V3 版本.

V2:

```
input: (batch_size, history time steps, features)

                 stride = 5
input -> expand features -> BN -> LSTM -> BN -> Dense(linear)
```

V3:

```
input: (batch_size, history time steps, features)

                stride = 5
        +-> expand features -> BN -> GRU -> BN -+
input --|       stride = 10                     |- concat -> Dense(linear)
        +-> expand features -> BN -> GRU -> BN -+
```

(BN: batch normalization)

version: 0.0.7

author: Congyu Wang

date: 2021-07-29

该module定义了计算不同时间序列特征的层，工程上使用tensorflow
进行高度向量化的计算，训练时较高效。
"""
import tensorflow as _tf
import tensorflow.keras.layers as _tfl
from tensorflow.keras.layers import Layer as _Layer
from tensorflow.keras.initializers import Initializer as _Initializer
from tensorflow.keras import Model as _Model
from .metrics import UpDownAccuracy as _UpDownAccuracy
from abc import ABC as _ABC
from abc import abstractmethod as _abstractmethod

if not "2.3.0" <= _tf.__version__:
    print(f"requires tensorflow version >= 2.3.0, "
          f"current version {_tf.__version__}")
    exit(1)

__all__ = ["Std",
           "Return",
           "Correlation",
           "LinearDecay",
           "Covariance",
           "ZScore",
           "FeatureExpansion",
           "AlphaNetV2",
           "AlphaNetV3",
           "AlphaNetV4",
           "load_model"]


class _StrideLayer(_Layer, _ABC):
    """计算每个stride的统计值的基类."""

    def __init__(self, stride=10, **kwargs):
        """计算每个stride的统计值的基类.

        Args:
            stride (int): time steps需要是stride的整数倍

        """
        if stride <= 1:
            raise ValueError("Illegal Argument: stride should be "
                             "greater than 1")
        super(_StrideLayer, self).__init__(**kwargs)
        self.stride = stride
        self.out_shape = None
        self.intermediate_shape = None

    def build(self, input_shape):
        """构建该层，计算维度信息."""
        (features,
         output_length) = __get_dimensions__(input_shape, self.stride)
        self.out_shape = [-1, output_length, features]
        self.intermediate_shape = [-1, self.stride, features]

    def get_config(self):
        """获取参数，保存模型需要的函数."""
        config = super().get_config().copy()
        config.update({'stride': self.stride})
        return config


class Std(_StrideLayer):
    """计算每个序列各stride的标准差.

    Notes:
        计算每个feature各个stride的standard deviation

    """

    def call(self, inputs, *args, **kwargs):
        """函数主逻辑实现部分.

        Args:
            inputs (tensor): 输入dimension为(batch_size, time_steps, features)

        Returns:
            dimension 为(batch_size, time_steps / stride, features)

        """
        strides = _tf.reshape(inputs, self.intermediate_shape)

        # compute standard deviations for each stride
        std = _tf.math.reduce_std(strides, axis=-2)
        return _tf.reshape(std, self.out_shape)


class ZScore(_StrideLayer):
    """计算每个序列各stride的均值除以其标准差.

    Notes:
        并非严格意义上的z-score,
        计算公式为每个feature各个stride的mean除以各自的standard deviation

    """

    def call(self, inputs, *args, **kwargs):
        """函数主逻辑实现部分.

        Args:
            inputs (tensor): 输入dimension为(batch_size, time_steps, features)

        Returns:
            dimension 为(batch_size, time_steps / stride, features)

        """
        strides = _tf.reshape(inputs, self.intermediate_shape)

        # compute standard deviations for each stride
        std = _tf.math.reduce_std(strides, axis=-2)

        # compute means for each stride
        means = _tf.math.reduce_mean(strides, axis=-2)

        # divide means by standard deviations for each stride
        z_score = _tf.math.divide_no_nan(means, std)
        return _tf.reshape(z_score, self.out_shape)


class LinearDecay(_StrideLayer):
    """计算每个序列各stride的线性衰减加权平均.

    Notes:
        以线性衰减为权重，计算每个feature各个stride的均值：
        如stride为10，则某feature该stride的权重为(1, 2, 3, 4, 5, 6, 7, 8, 9, 10)

    """

    def call(self, inputs, *args, **kwargs):
        """函数主逻辑实现部分.

        Args:
            inputs (tensor): 输入dimension为(batch_size, time_steps, features)

        Returns:
            dimension 为(batch_size, time_steps / stride, features)

        """
        # get linear decay kernel
        single_kernel = _tf.linspace(1.0, self.stride, num=self.stride)
        kernel = _tf.repeat(single_kernel, self.intermediate_shape[2])
        kernel = kernel / _tf.reduce_sum(single_kernel)

        # reshape tensors into:
        # (bash_size * (time_steps / stride), stride, features)
        kernel = _tf.reshape(kernel, self.intermediate_shape[1:])
        inputs = _tf.reshape(inputs, self.intermediate_shape)

        # broadcasting kernel to inputs batch dimension
        linear_decay = _tf.reduce_sum(kernel * inputs, axis=1)
        linear_decay = _tf.reshape(linear_decay, self.out_shape)
        return linear_decay


class Return(_Layer):
    """计算每个序列各stride的回报率.

    Notes:
        计算公式为每个stride最后一个数除以第一个数再减去一

    """

    def __init__(self, stride=10, **kwargs):
        """回报率.

        Args:
            stride (int): time steps需要是stride的整数倍

        """
        if stride <= 1:
            raise ValueError("Illegal Argument: stride should be "
                             "greater than 1")
        super(Return, self).__init__(**kwargs)
        self.stride = stride

    def build(self, input_shape):
        """构建该层，计算维度信息."""
        time_steps = input_shape[1]
        if time_steps % self.stride != 0:
            raise ValueError("Error, time_steps 应该是 stride的整数倍")

    def call(self, inputs, *args, **kwargs):
        """函数主逻辑实现部分.

        Args:
            inputs (tensor): 输入dimension为(batch_size, time_steps, features)

        Returns:
            dimension 为(batch_size, time_steps / stride, features)

        """
        # get the endings of each strides as numerators
        numerators = inputs[:, (self.stride - 1)::self.stride, :]

        # get the beginnings of each strides as denominators
        denominators = inputs[:, 0::self.stride, :]

        return _tf.math.divide_no_nan(numerators, denominators) - 1.0

    def get_config(self):
        """获取参数，保存模型需要的函数."""
        config = super().get_config().copy()
        config.update({'stride': self.stride})
        return config


class _OuterProductLayer(_Layer, _ABC):

    def __init__(self, stride=10, **kwargs):
        """外乘类的扩张层.

        Args:
            stride (int): time steps需要是stride的整数倍

        """
        if stride <= 1:
            raise ValueError("Illegal Argument: stride should be "
                             "greater than 1")
        super(_OuterProductLayer, self).__init__(**kwargs)
        self.stride = stride
        self.intermediate_shape = None
        self.out_shape = None
        self.lower_mask = None

    def build(self, input_shape):
        """构建该层，计算维度信息."""
        (features,
         output_length) = __get_dimensions__(input_shape, self.stride)
        self.intermediate_shape = (-1, self.stride, features)
        output_features = int(features * (features - 1) / 2)
        self.out_shape = (-1, output_length, output_features)
        self.lower_mask = _LowerNoDiagonalMask()((features, features))

    def get_config(self):
        """获取参数，保存模型需要的函数."""
        config = super().get_config().copy()
        config.update({'stride': self.stride})
        return config

    @_abstractmethod
    def call(self, inputs, *args, **kwargs):
        """逻辑实现部分."""
        ...


class Covariance(_OuterProductLayer):
    """计算每个stride各时间序列片段的covariance.

    Notes:
        计算每个stride每两个feature之间的covariance大小，
        输出feature数量为features * (features - 1) / 2

    """

    def call(self, inputs, *args, **kwargs):
        """函数主逻辑实现部分.

        Args:
            inputs (tensor): 输入dimension为(batch_size, time_steps, features)

        Returns:
            dimension 为(batch_size, time_steps / stride,
            features * (features - 1) / 2)

        """
        # compute means for each stride
        means = _tf.nn.avg_pool(inputs,
                                ksize=self.stride,
                                strides=self.stride,
                                padding="VALID")

        # subtract means for each stride
        means_broadcast = _tf.repeat(means, self.stride, axis=1)
        means_subtracted = _tf.subtract(inputs, means_broadcast)
        means_subtracted = _tf.reshape(means_subtracted,
                                       self.intermediate_shape)

        # compute covariance matrix
        covariance_matrix = _tf.einsum("ijk,ijm->ikm",
                                       means_subtracted,
                                       means_subtracted)
        covariance_matrix = covariance_matrix / (self.stride - 1)

        # get the lower part of the covariance matrix
        # without the diagonal elements
        covariances = _tf.boolean_mask(covariance_matrix,
                                       self.lower_mask,
                                       axis=1)
        covariances = _tf.reshape(covariances, self.out_shape)
        return covariances


class Correlation(_OuterProductLayer):
    """计算每个stride各时间序列的相关系数.

    Notes:
        计算每个stride每两个feature之间的correlation coefficient，
        输出feature数量为features * (features - 1) / 2

    """

    def call(self, inputs, *args, **kwargs):
        """函数主逻辑实现部分.

        Args:
            inputs (tensor): 输入dimension为(batch_size, time_steps, features)

        Returns:
            dimension 为(batch_size, time_steps / stride,
            features * (features - 1) / 2)

        """
        # compute means for each stride
        means = _tf.nn.avg_pool(inputs,
                                ksize=self.stride,
                                strides=self.stride,
                                padding="VALID")

        # subtract means for each stride
        means_broadcast = _tf.repeat(means, self.stride, axis=1)
        means_subtracted = _tf.subtract(inputs, means_broadcast)
        means_subtracted = _tf.reshape(means_subtracted,
                                       self.intermediate_shape)

        # compute standard deviations for each strides
        squared_diff = _tf.square(means_subtracted)
        mean_squared_error = _tf.reduce_mean(squared_diff, axis=1)
        std = _tf.sqrt(mean_squared_error)

        # get denominator of correlation matrix
        denominator_matrix = _tf.einsum("ik,im->ikm", std, std)

        # compute covariance matrix
        covariance_matrix = _tf.einsum("ijk,ijm->ikm",
                                       means_subtracted,
                                       means_subtracted)
        covariance_matrix = covariance_matrix / self.stride

        # take the lower triangle of each matrix without diagonal
        covariances = _tf.boolean_mask(covariance_matrix,
                                       self.lower_mask,
                                       axis=1)
        denominators = _tf.boolean_mask(denominator_matrix,
                                        self.lower_mask,
                                        axis=1)
        correlations = _tf.math.divide_no_nan(covariances, denominators)
        correlations = _tf.reshape(correlations, self.out_shape)
        return correlations


class FeatureExpansion(_Layer):
    """计算时间序列特征扩张层，汇总6个计算层.

    Notes:
        该层扩张时间序列的feature数量，并通过stride缩短时间序列长度，
        其包括一下一些feature:

            - standard deviation

            - mean / standard deviation

            - linear decay average

            - return of each stride

            - covariance of each two features for each stride

            - correlation coefficient of each two features for each stride

    """

    def __init__(self, stride=10, **kwargs):
        """时间序列特征扩张.

        Args:
            stride (int): time steps需要是stride的整数倍

        """
        if type(stride) is not int or stride <= 1:
            raise ValueError("Illegal Argument: stride should be an integer "
                             "greater than 1")
        super(FeatureExpansion, self).__init__(**kwargs)
        self.stride = stride
        self.std = _tf.function(Std(stride=self.stride))
        self.z_score = _tf.function(ZScore(stride=self.stride))
        self.linear_decay = _tf.function(LinearDecay(stride=self.stride))
        self.return_ = _tf.function(Return(stride=self.stride))
        self.covariance = _tf.function(Covariance(stride=self.stride))
        self.correlation = _tf.function(Correlation(stride=self.stride))

    def call(self, inputs, *args, **kwargs):
        """函数主逻辑实现部分.

        Args:
            inputs (tensor): 输入dimension为(batch_size, time_steps, features)

        Returns:
            dimension 为(batch_size, time_steps / stride,
            features * (features + 3))

        """
        std_output = self.std(inputs)
        z_score_output = self.z_score(inputs)
        decay_linear_output = self.linear_decay(inputs)
        return_output = self.return_(inputs)
        covariance_output = self.covariance(inputs)
        correlation_output = self.correlation(inputs)
        return _tf.concat([std_output,
                           z_score_output,
                           decay_linear_output,
                           return_output,
                           covariance_output,
                           correlation_output], axis=2)

    def get_config(self):
        """获取参数，保存模型需要的函数."""
        config = super().get_config().copy()
        config.update({'stride': self.stride})
        return config


class AlphaNetV2(_Model):
    """神经网络模型，继承``keras.Model``类.

    alpha net v2版本模型.

    Notes:
        复现华泰金工 alpha net V2 版本

        ``input: (batch_size, history time steps, features)``

    """

    def __init__(self,
                 dropout=0.0,
                 l2=0.001,
                 stride=10,
                 classification=False,
                 categories=0,
                 *args,
                 **kwargs):
        """Alpha net v3.

        Notes:
            alpha net v2 版本的全tensorflow实现，结构详见代码展开

        Args:
            dropout: 跟在特征扩张以及Batch Normalization之后的dropout，默认无dropout
            l2: 输出层的l2-regularization参数

        """
        super(AlphaNetV2, self).__init__(*args, **kwargs)
        self.l2 = l2
        self.dropout = dropout
        self.stride = stride
        self.expanded = FeatureExpansion(stride=self.stride)
        self.normalized = _tfl.BatchNormalization()
        self.dropout = _tfl.Dropout(self.dropout)
        self.lstm = _tfl.LSTM(units=30)
        self.normalized_2 = _tfl.BatchNormalization()
        self.regularizer = _tf.keras.regularizers.l2(self.l2)
        if classification:
            if categories < 1:
                raise ValueError("categories should be at least 1")
            elif categories == 1:
                self.outputs = _tfl.Dense(1, activation="sigmoid",
                                          kernel_initializer="truncated_normal",
                                          kernel_regularizer=self.regularizer)
            else:
                self.outputs = _tfl.Dense(categories, activation="softmax",
                                          kernel_initializer="truncated_normal",
                                          kernel_regularizer=self.regularizer)
        else:
            self.outputs = _tfl.Dense(1, activation="linear",
                                      kernel_initializer="truncated_normal",
                                      kernel_regularizer=self.regularizer)

    @_tf.function
    def call(self, inputs, training=None, mask=None):
        """计算逻辑实现."""
        expanded = self.expanded(inputs)
        normalized = self.normalized(expanded, training=training)
        lstm = self.lstm(normalized)
        normalized2 = self.normalized_2(lstm, training=training)
        dropout = self.dropout(normalized2, training=training)
        output = self.outputs(dropout)
        return output

    def compile(self,
                optimizer=_tf.keras.optimizers.Adam(0.0001),
                loss="MSE",
                metrics=None,
                loss_weights=None,
                weighted_metrics=None,
                run_eagerly=None,
                **kwargs):
        """设置优化器、loss、metric等."""
        super().compile(optimizer=optimizer,
                        loss=loss,
                        metrics=metrics,
                        loss_weights=loss_weights,
                        weighted_metrics=weighted_metrics,
                        run_eagerly=run_eagerly)

    def get_config(self):
        """获取参数，保存模型需要的函数."""
        config = super().get_config().copy()
        config.update({'dropout': self.dropout,
                       'l2': self.l2,
                       'stride': self.stride})
        return config


class AlphaNetV3(_Model):
    """神经网络模型，继承``keras.Model``类.

    alpha net v3版本模型.

    Notes:
        复现华泰金工 alpha net V3 版本

        ``input: (batch_size, history time steps, features)``

    """

    def __init__(self,
                 dropout=0.0,
                 l2=0.001,
                 classification=False,
                 categories=0,
                 recurrent_unit="GRU",
                 hidden_units=30,
                 *args,
                 **kwargs):
        """Alpha net v3.

        Notes:
            alpha net v3 版本的全tensorflow实现，结构详见代码展开

        Args:
            dropout: 跟在特征扩张以及Batch Normalization之后的dropout，默认无dropout
            l2: 输出层的l2-regularization参数
            classification: 是否为分类问题
            categories: 分类问题的类别数量
            recurrent_unit (str): 该参数可以为"GRU"或"LSTM"

        """
        super(AlphaNetV3, self).__init__(*args, **kwargs)
        self.l2 = l2
        self.dropout = dropout
        self.expanded10 = FeatureExpansion(stride=10)
        self.expanded5 = FeatureExpansion(stride=5)
        self.normalized10 = _tfl.BatchNormalization()
        self.normalized5 = _tfl.BatchNormalization()
        self.dropout_layer = _tfl.Dropout(self.dropout)
        if recurrent_unit == "GRU":
            self.recurrent10 = _tfl.GRU(units=hidden_units)
            self.recurrent5 = _tfl.GRU(units=hidden_units)
        elif recurrent_unit == "LSTM":
            self.recurrent10 = _tfl.LSTM(units=hidden_units)
            self.recurrent5 = _tfl.LSTM(units=hidden_units)
        else:
            raise ValueError("Unknown recurrent_unit")
        self.normalized10_2 = _tfl.BatchNormalization()
        self.normalized5_2 = _tfl.BatchNormalization()
        self.concat = _tfl.Concatenate(axis=-1)
        self.regularizer = _tf.keras.regularizers.l2(self.l2)
        if classification:
            if categories < 1:
                raise ValueError("categories should be at least 1")
            elif categories == 1:
                self.outputs = _tfl.Dense(1, activation="sigmoid",
                                          kernel_initializer="truncated_normal",
                                          kernel_regularizer=self.regularizer)
            else:
                self.outputs = _tfl.Dense(categories, activation="softmax",
                                          kernel_initializer="truncated_normal",
                                          kernel_regularizer=self.regularizer)
        else:
            self.outputs = _tfl.Dense(1, activation="linear",
                                      kernel_initializer="truncated_normal",
                                      kernel_regularizer=self.regularizer)

    @_tf.function
    def call(self, inputs, training=None, mask=None):
        """计算逻辑实现."""
        expanded10 = self.expanded10(inputs)
        expanded5 = self.expanded5(inputs)
        normalized10 = self.normalized10(expanded10, training=training)
        normalized5 = self.normalized5(expanded5, training=training)
        recurrent10 = self.recurrent10(normalized10)
        recurrent5 = self.recurrent5(normalized5)
        normalized10_2 = self.normalized10_2(recurrent10, training=training)
        normalized5_2 = self.normalized5_2(recurrent5, training=training)
        concat = self.concat([normalized10_2, normalized5_2])
        dropout = self.dropout_layer(concat, training=training)
        output = self.outputs(dropout)
        return output

    def compile(self,
                optimizer=_tf.keras.optimizers.Adam(0.0001),
                loss="MSE",
                metrics=None,
                loss_weights=None,
                weighted_metrics=None,
                run_eagerly=None,
                **kwargs):
        """设置优化器、loss、metric等."""
        super().compile(optimizer=optimizer,
                        loss=loss,
                        metrics=metrics,
                        loss_weights=loss_weights,
                        weighted_metrics=weighted_metrics,
                        run_eagerly=run_eagerly)

    def get_config(self):
        """获取参数，保存模型需要的函数."""
        config = super().get_config().copy()
        config.update({'dropout': self.dropout,
                       'l2': self.l2})
        return config


class AlphaNetV4(_Model):
    """神经网络模型，继承``keras.Model``类.

    Notes:
        ``input: (batch_size, history time steps, features)``

    """

    def __init__(self,
                 dropout=0.0,
                 l2=0.001,
                 classification=False,
                 categories=0,
                 recurrent_unit="GRU",
                 hidden_units=30,
                 *args,
                 **kwargs):
        """Alpha net v4.

        Notes:
            去掉了batch normalization的模型，
            训练需要使用data模块的normalization
            或其他自定义normalization.

        Args:
            dropout: 跟在特征扩张以及Batch Normalization之后的dropout，默认无dropout
            l2: 输出层的l2-regularization参数
            classification: 是否为分类问题
            categories: 分类问题的类别数量
            recurrent_unit (str): 该参数可以为"GRU"或"LSTM"

        """
        super(AlphaNetV4, self).__init__(*args, **kwargs)
        self.l2 = l2
        self.dropout = dropout
        self.expanded10 = FeatureExpansion(stride=10)
        self.expanded5 = FeatureExpansion(stride=5)
        self.dropout_layer = _tfl.Dropout(self.dropout)
        if recurrent_unit == "GRU":
            self.recurrent10 = _tfl.GRU(units=hidden_units)
            self.recurrent5 = _tfl.GRU(units=hidden_units)
        elif recurrent_unit == "LSTM":
            self.recurrent10 = _tfl.LSTM(units=hidden_units)
            self.recurrent5 = _tfl.LSTM(units=hidden_units)
        else:
            raise ValueError("Unknown recurrent_unit")
        self.concat = _tfl.Concatenate(axis=-1)
        self.regularizer = _tf.keras.regularizers.l2(self.l2)
        if classification:
            if categories < 1:
                raise ValueError("categories should be at least 1")
            elif categories == 1:
                self.outputs = _tfl.Dense(1, activation="sigmoid",
                                          kernel_initializer="truncated_normal",
                                          kernel_regularizer=self.regularizer)
            else:
                self.outputs = _tfl.Dense(categories, activation="softmax",
                                          kernel_initializer="truncated_normal",
                                          kernel_regularizer=self.regularizer)
        else:
            self.outputs = _tfl.Dense(1, activation="linear",
                                      kernel_initializer="truncated_normal",
                                      kernel_regularizer=self.regularizer)

    @_tf.function
    def call(self, inputs, training=None, mask=None):
        """计算逻辑实现."""
        expanded10 = self.expanded10(inputs)
        expanded5 = self.expanded5(inputs)
        recurrent10 = self.recurrent10(expanded10)
        recurrent5 = self.recurrent5(expanded5)
        concat = self.concat([recurrent10, recurrent5])
        dropout = self.dropout_layer(concat, training=training)
        output = self.outputs(dropout)
        return output

    def compile(self,
                optimizer=_tf.keras.optimizers.Adam(0.0001),
                loss="MSE",
                metrics=None,
                loss_weights=None,
                weighted_metrics=None,
                run_eagerly=None,
                **kwargs):
        """设置优化器、loss、metric等."""
        super().compile(optimizer=optimizer,
                        loss=loss,
                        metrics=metrics,
                        loss_weights=loss_weights,
                        weighted_metrics=weighted_metrics,
                        run_eagerly=run_eagerly)

    def get_config(self):
        """获取参数，保存模型需要的函数."""
        config = super().get_config().copy()
        config.update({'dropout': self.dropout,
                       'l2': self.l2})
        return config


def load_model(filepath,
               custom_objects: dict = None,
               compile: bool = True,
               options=None):
    """用于读取已存储的模型，可识别自定义metric: UpDownAccuracy.

    Notes:
        包装``tf.keras``的``load_model``函数，添加``UpDownAccuracy``.

    Args:
        filepath: 文件路径:
            - String or `pathlib.Path` object, path to the saved model
            - `h5py.File` object from which to load the model
        custom_objects: 自定义类的识别，从类或函数名到类或函数的映射字典.
        compile: Boolean, 是否compile model.
        options: 其他 `tf.saved_model.LoadOptions`.

    Returns:
        Keras model instance.

    Raises:
        ImportError: if loading from an hdf5 file and h5py is not available.
        IOError: In case of an invalid savefile
    """
    object_dict = {"UpDownAccuracy": _UpDownAccuracy}
    if custom_objects is not None:
        object_dict.update(custom_objects)
    return _tf.keras.models.load_model(filepath,
                                       custom_objects=object_dict,
                                       compile=compile,
                                       options=options)


class _LowerNoDiagonalMask(_Initializer):
    """获取不含对角元素的矩阵下三角mask.

    Notes:
        Provide a mask giving the lower triangular of a matrix
        without diagonal elements.

    """

    def __init__(self):
        super(_LowerNoDiagonalMask, self).__init__()

    def __call__(self, shape, **kwargs):
        """计算逻辑."""
        ones = _tf.ones(shape)
        mask_lower = _tf.linalg.band_part(ones, -1, 0)
        mask_diag = _tf.linalg.band_part(ones, 0, 0)
        # lower triangle removing the diagonal elements
        mask = _tf.cast(mask_lower - mask_diag, dtype=_tf.bool)
        return mask


def __get_dimensions__(input_shape, stride):
    """计算相关维度长度.

    Notes:
        output_length = 原来的时间长度 / stride的长度

    Args:
        input_shape: pass the inputs of layer to the function
        stride (int): the stride of the custom layer

    Returns:
        (features, output_length)

    Raises:
        ValueError: 如果历史长度不是stride的整数倍

    """
    if type(stride) is not int or stride <= 1:
        raise ValueError("Illegal Argument: stride should be an integer "
                         "greater than 1")
    time_steps = input_shape[1]
    features = input_shape[2]
    output_length = time_steps // stride

    if time_steps % stride != 0:
        raise ValueError("Error, time_steps 应该是 stride的整数倍")

    return features, output_length



"""训练中的辅助准确率信息.

该module包含涨跌精度信息的计算类。
使用到该类的模型在保存后重新读取时需要添加``custom_objects``参数，
或者使用``alphanet.load_model()``函数。

"""
import tensorflow as _tf


__all__ = ["UpDownAccuracy"]


class UpDownAccuracy(_tf.keras.metrics.Metric):
    """通过对return的预测来计算涨跌准确率."""

    def __init__(self, name='up_down_accuracy', **kwargs):
        """涨跌准确率."""
        super(UpDownAccuracy, self).__init__(name=name, **kwargs)
        self.up_down_correct_count = self.add_weight(name='ud_count',
                                                     initializer='zeros',
                                                     shape=(),
                                                     dtype=_tf.float32)
        self.length = self.add_weight(name='len',
                                      initializer='zeros',
                                      shape=(),
                                      dtype=_tf.float32)

    def update_state(self, y_true, y_pred, sample_weight=None):
        """加入新的预测更新精度值."""
        y_true = _tf.cast(y_true > 0.0, _tf.float32)
        y_pred = _tf.cast(y_pred > 0.0, _tf.float32)
        length = _tf.cast(len(y_true), _tf.float32)
        correct_count = length - _tf.reduce_sum(_tf.abs(y_true - y_pred))

        self.length.assign_add(length)
        self.up_down_correct_count.assign_add(correct_count)

    def result(self):
        """获取结果."""
        if self.length == 0.0:
            return 0.0
        return self.up_down_correct_count / self.length

    def reset_state(self):
        """在train、val的epoch末尾重置精度."""
        self.up_down_correct_count.assign(0.0)
        self.length.assign(0.0)



"""多维多时间序列神经网络滚动训练的数据工具.

version: 0.0.19

author: Congyu Wang

date: 2021-08-26
"""
import numpy as np
import tensorflow as _tf
import numpy as _np
from numba import njit
from typing import List as _List

__all__ = ["TimeSeriesData", "TrainValData"]


class TimeSeriesData:
    """单个时间序列信息.

    Notes:
        用于储存个股的数据信息及预测label，全部使用numpy，日期格式为整数: ``YYYYMMDD``。
        数据分三个部分：时间，数据，标签，第一个维度都是时间，数据的第二个维度为特征。

    """

    def __init__(self,
                 dates: _np.ndarray,
                 data: _np.ndarray,
                 labels: _np.ndarray):
        """储存个股的数据信息及预测目标，全部使用numpy，日期格式为整数: ``YYYYMMDD``.

        Args:
            dates: 日期列, 1D ``numpy.ndarray``, 整数
            data: 训练输入的X，2D ``numpy.ndarray``, (日期长度 x 特征数量)
            labels: 训练标签Y, 1D ``numpy.ndarray``, 长度与dates相同。
                如果为分类问题则是2D, (日期长度 x 类别数量)

        """
        # 检查参数类型
        if (not type(dates) is _np.ndarray or
                not type(data) is _np.ndarray or
                not type(labels) is _np.ndarray):
            raise ValueError("Data should be numpy arrays")
        # 检查日期、数据、标签长度是否一致
        if len(dates) != len(data) or len(dates) != len(labels):
            raise ValueError("Bad data shape")
        # 检查维度是否正确
        if dates.ndim != 1 or data.ndim != 2 or not 1 <= labels.ndim <= 2:
            raise ValueError("Wrong dimensions")
        self.dates = dates.astype(_np.int32)
        self.data = data
        self.labels = labels


class TrainValData:
    """根据训练天数、验证天数、样本历史长度、训练起点生成不同训练阶段的数据."""

    def __init__(self,
                 time_series_list: _List[TimeSeriesData],
                 train_length: int = 1200,
                 validate_length: int = 300,
                 history_length: int = 30,
                 train_val_gap: int = 10,
                 sample_step: int = 2,
                 fill_na: _np.float = _np.NaN,
                 normalize: bool = False):
        """用于获取不同阶段的训练集和验证集.

        Notes:
            ``time_series_list``储存全部的时间序列信息，
            其中每支股票序列为一个单独``TimeSeriesData``，
            完整数据为``List[TimeSeriesData]``类型。

            此外需要提供训练集总交易天数(``train_length``)、
            验证集总交易天数(``validate_length``)、
            单个样本用到的历史长度(``history_length``)、
            采样步进大小(``sample_step``)。

            使用方法为：通过get(start_date)方法获取从start_date
            开始的训练机和验证集。通过逐渐增大start_date训练多个模型回测。

            ``train_val_gap``参数为验证集第一天与训练集最后一天中间间隔的天数，
            如果是相临，则train_val_gap = 0。设置该参数的目的如下：

            如果希望预测未来十天的累计收益，则预测时用到的输入数据为最近的历史数据来预测
            未来十天的累计收益，即用t(-history)到t(0)的数据来预测t(1)到t(11)的累计收益
            而训练时因为要用到十天累计收益做标签，最近的一个十天累计收益是从t(-10)到t(0)，
            用到的历史数据则必须是t(-history-11)到t(-11)的数据。
            而validation时，如果第一个预测点是t(1)(明天收盘价)至t(11)的累计收益，
            则与最后一个训练的数据即：t(-10)至t(0)之间间隔了10天，
            使用``train_val_gap=10``。

            可选项为fill_na，缺失数据填充值，默认为np.Na
            训练时跳过所有有缺失数据的样本。

        Args:
            time_series_list: TimeSeriesData 列表
            train_length: 训练集天数
            validate_length: 验证集天数
            history_length: 每个样本的历史天数
            train_val_gap: 训练集与验证集的间隔
            sample_step: 采样sample时步进的天数
            fill_na: 默认填充为np.NaN，训练时会跳过有确实数据的样本
            normalize: 是否对非率值做每个历史片段的max/min标准化

        """
        # 检查参数类型
        if type(time_series_list) is not list:
            raise ValueError("time_series_list should be a list")
        # 不允许空列表
        if len(time_series_list) == 0:
            raise ValueError("Empty time_series_list")
        # 检查列表元素类型
        for t in time_series_list:
            if type(t) is not TimeSeriesData:
                raise ValueError("time_series_data should be a list "
                                 "of TimeSeriesData objects")
        # 检查参数数值
        if (type(history_length) is not int or
                type(validate_length) is not int or
                type(sample_step) is not int or
                type(train_length) is not int or
                type(train_val_gap) is not int or
                history_length < 1 or
                validate_length < 1 or
                sample_step < 1 or
                train_val_gap < 0 or
                train_length < history_length):
            raise ValueError("bad arguments")

        if type(fill_na) is not _np.float:
            raise ValueError("fill_na should be numpy float")

        # 确保数据特征数量一致
        self.__feature_counts = time_series_list[0].data.shape[1]
        for series in time_series_list:
            if series.data.shape[1] != self.__feature_counts:
                raise ValueError("time series do not have "
                                 "the same number of features")

        # 确保标签维度一致
        label_dims = time_series_list[0].labels.ndim
        for series in time_series_list:
            if series.labels.ndim != label_dims:
                raise ValueError("time labels do not have "
                                 "the same number of dimensions")

        # 标签类别数量
        class_num = 0
        if label_dims == 2:
            class_num = time_series_list[0].labels.shape[1]
            for series in time_series_list:
                if series.labels.shape[1] != class_num:
                    raise ValueError("time series labels do not have "
                                     "the same number of classes")

        self.__class_num = class_num

        # 获取日期列表（所有时间序列日期的并集）
        self.__distinct_dates = _np.unique([date for stock in time_series_list
                                            for date in stock.dates])
        self.__distinct_dates.sort()

        # 聚合数据为(序列(股票), 时间, 特征数量)的张量，缺失数据为np.NaN
        # 标签维度为(序列(股票), 时间)
        self.__data = _np.empty((len(time_series_list),
                                 len(self.__distinct_dates),
                                 self.__feature_counts))
        if self.__class_num == 0:
            self.__labels = _np.empty((len(time_series_list),
                                       len(self.__distinct_dates)))
        else:
            self.__labels = _np.empty((len(time_series_list),
                                       len(self.__distinct_dates),
                                       self.__class_num))

        self.__series_date_matrix = _np.empty((len(time_series_list),
                                               len(self.__distinct_dates), 2),
                                              dtype=_np.int)
        self.__data[:] = fill_na
        self.__labels[:] = fill_na

        # 根据日期序列的位置向张量填充数据
        dates_positions = {date: index
                           for index, date in enumerate(self.__distinct_dates)}
        dates_position_mapper = _np.vectorize(lambda d: dates_positions[d])
        for i, series in enumerate(time_series_list):
            # 找到该序列series.dates日期在日期列表中的位置
            # 将第i个序列填充至tensor的第i行
            position_index = dates_position_mapper(series.dates)
            self.__data[i, position_index, :] = series.data
            if self.__class_num == 0:
                self.__labels[i, position_index] = series.labels
            else:
                self.__labels[i, position_index, :] = series.labels
            self.__series_date_matrix[i, position_index, 0] = series.dates
            self.__series_date_matrix[i, position_index, 1] = i

        self.__train_length = train_length
        self.__validate_length = validate_length
        self.__history_length = history_length
        self.__sample_step = sample_step
        self.__train_val_gap = train_val_gap
        self.__normalize = normalize

    def get(self,
            start_date: int,
            order="by_date",
            validate_only=False,
            validate_length=None,
            normalize=False):
        """获取从某天开始的训练集和验证集.

        Notes:
            根据设定的训练集天数以及验证集天数，从start_date开始获取正确的
            训练集以及验证集，以及他们各自的日期范围信息(该信息以字典形式返回)。

            需要注意:
            训练集的的开始和结束是指其data, label时间范围并集，而
            验证集的开始和结束则只是指其label的时间范围。
            验证集的输入数据可以与训练集重叠，只要其标签数据的包含的时间范围
            与训练集数据包含的时间范围没有交集即可。

            具体时间信息参考函数返回的最后一个元素，是一个包含时间信息的``dict``。

        Args:
            start_date: 该轮训练开始日期，整数``YYYYMMDD``
            order: 有三种顺序 ``shuffle``, ``by_date``, ``by_series``。
                分别为随机打乱股票和时间，按时间顺序优先，按股票顺序优先，默认by_date。
            validate_only: 如果设置为True，则只返回validate set
                和训练集、验证集时间信息。可以用于训练后的分析。
            validate_length (int): override class validate_length
            normalize (bool): override class normalize

        Returns:
            如果``validate_only=False``，返回训练集、验证集、日期信息：
            (train, val, dates_info(dict))。
            如果为``True``，则返回验证集、日期信息。

        Raises:
            ValueError: 日期范围超出最大日期会报ValueError。

        """
        if validate_length is not None:
            if type(validate_length) is not int:
                raise ValueError("`validate_length` must be an integer")
            if validate_length < 1:
                raise ValueError("`validate_length` should be at least 1")
        else:
            validate_length = self.__validate_length
        if not normalize:
            normalize = self.__normalize
        return self.__get_in_memory__(start_date,
                                      order,
                                      validate_only,
                                      validate_length,
                                      normalize)

    def __get_in_memory__(self,
                          start_date,
                          order="by_date",
                          validate_only=False,
                          validate_length=None,
                          normalize=False):
        """使用显存生成历史数据.

        使用tensorflow from_tensor_slices，通过传递完整的tensor进行训练，
        股票数量大时，需要较大内存

        """
        # 获取用于构建训练集、验证集的相关信息
        kwargs = {"start_date": start_date, "order": order}
        if validate_length is not None:
            kwargs.update({"validate_length": validate_length})
        train_args, val_args, dates_info = self.__get_period_info__(**kwargs)
        train_args = (*train_args, normalize)
        val_args = (*val_args, normalize)
        # 将输入的数据、标签片段转化为单个sample包含history日期长度的历史信息
        (val_x,
         val_y,
         val_dates_series) = __full_tensor_generation__(*val_args)
        # 转化为tensorflow DataSet
        val = _tf.data.Dataset.from_tensor_slices((val_x, val_y))
        val_dates_list = val_dates_series[:, 0].tolist()
        val_series_list = val_dates_series[:, 1].tolist()
        dates_info["validation"]["dates_list"] = val_dates_list
        dates_info["validation"]["series_list"] = val_series_list

        if validate_only:
            return val, dates_info

        (train_x,
         train_y,
         train_dates_series) = __full_tensor_generation__(*train_args)
        train = _tf.data.Dataset.from_tensor_slices((train_x, train_y))
        train_dates_list = train_dates_series[:, 0].tolist()
        train_series_list = train_dates_series[:, 1].tolist()
        dates_info["training"]["dates_list"] = train_dates_list
        dates_info["training"]["series_list"] = train_series_list
        return train, val, dates_info

    def __get_period_info__(self, start_date, order="by_date",
                            validate_length=None):
        """根据开始时间计算用于构建训练集、验证集的相关信息."""
        if type(start_date) is not int:
            raise ValueError("start date should be an integer YYYYMMDD")

        # 找到大于等于start_date的最小日期
        after_start_date = self.__distinct_dates >= start_date
        first_date = _np.min(self.__distinct_dates[after_start_date])

        # 查看剩余日期数量是否大于等于训练集验证集总长度
        if _np.sum(after_start_date) < (self.__train_length +
                                        validate_length +
                                        self.__train_val_gap):
            raise ValueError("date range exceeded end of dates")

        # 计算各个时间节点在时间列表中的位置
        # 训练集开始位置(data的开始位置)
        train_start_index = __first_index__(self.__distinct_dates, first_date)
        # 训练集结束位置(不包含)
        train_end_index = train_start_index + self.__train_length
        # 验证集开始位置(data的开始位置)
        val_start_index = (train_end_index -
                           self.__history_length +
                           self.__train_val_gap + 1)
        # 验证集结束位置(不包含)
        val_end_index = (train_end_index +
                         validate_length +
                         self.__train_val_gap)

        # 根据各个数据集的开始结束位置以及训练数据的顺序选项，获取构建数据的参数
        train_args = self.__get_generator_args__(train_start_index,
                                                 train_end_index,
                                                 order=order)
        val_args = self.__get_generator_args__(val_start_index,
                                               val_end_index,
                                               order=order)
        dates_info = self.__dates_info__(train_start_index,
                                         val_start_index,
                                         train_args,
                                         val_args)

        return train_args, val_args, dates_info

    def __get_generator_args__(self, start_index, end_index, order="by_date"):
        """获取单个generator需要的数据片段.

        Notes:
            根据数据集的开始、结束位置以及的顺序选项，获取该训练集的数据、标签片段
            以及用于生成训练数据的(序列, 日期)pair列表的顺序信息(generation_list)。

            generation_list第一列为序列编号，第二列为日期。

            generation_list中的日期数字代表*每个历史数据片段*的第一个日期相对
            该数据集片段（start_index:end_index）的位置。

            注意：

                - 该处的日期列表不代表每个历史片段的结束位置

                - 也不是相对TrainValData类日期列表的位置
        """
        length = end_index - start_index
        data = self.__data[:, start_index:end_index, :]
        label = self.__labels[:, start_index:end_index]
        dates_series = self.__series_date_matrix[:, start_index:end_index, :]
        generation_list = [[series_i, t]
                           for t in range(0,
                                          length - self.__history_length + 1,
                                          self.__sample_step)
                           for series_i in range(len(data))]

        if order == "shuffle":
            generation_list = _np.array(generation_list)
            _np.random.shuffle(generation_list)
        elif order == "by_date":
            generation_list = _np.array(generation_list)
        elif order == "by_series":
            generation_list = sorted(generation_list, key=lambda k: k[0])
            generation_list = _np.array(generation_list)
        else:
            raise ValueError("wrong order argument, choose from `shuffle`, "
                             "`by_date`, and `by_series`")

        history_length = self.__history_length

        return data, label, generation_list, history_length, dates_series

    def __dates_info__(self,
                       train_start_index,
                       val_start_index,
                       train_args,
                       val_args):
        """根据生成数据的列表，计算用于显示的日期信息."""
        # 获取generation_list(生成数据的顺序信息)的时间列
        # 该时间列为相对数据片段开头的时间位置
        train_generation_list = train_args[2]
        val_generation_list = val_args[2]
        train_time_index = train_generation_list[:, 1]
        val_time_index = val_generation_list[:, 1]

        # 加上片段的开始位置，得到相对TrainValData类日期列表的位置
        train_time_index = train_time_index + train_start_index
        val_time_index = val_time_index + val_start_index

        # 训练集在日期列表中的开始位置
        training_beginning = _np.min(train_time_index)

        # 结束位置：加上历史长度减去一，获取最大日期位置(inclusive)
        training_ending = _np.max(train_time_index)
        training_ending += self.__history_length - 1

        # validation集每次取的都是某历史片段末尾的数据
        # 所以加上历史减去一
        validation_index = val_time_index + self.__history_length - 1
        validation_beginning = _np.min(validation_index)
        validation_ending = _np.max(validation_index)

        dates_info = {
            "training": {
                "start_date": int(self.__distinct_dates[training_beginning]),
                "end_date": int(self.__distinct_dates[training_ending]),
            },
            "validation": {
                "start_date": int(self.__distinct_dates[validation_beginning]),
                "end_date": int(self.__distinct_dates[validation_ending]),
            }
        }
        return dates_info


def __full_tensor_generation__(data,
                               label,
                               generation_list,
                               history,
                               dates_series,
                               normalize):
    """将输入的数据、标签片段转化为单个sample包含history日期长度的历史信息."""
    # 先将该数据片段的历史维度展开

    # 根据generation_list指定的series，日期，获取标签及数据片段
    total_time_length = data.shape[1]
    expanded = [__gather_2d__(data[:, i: (total_time_length + i
                                          - history + 1), :], generation_list)
                for i in range(history)]

    # date_all dimensions: (series * dates, features, history)
    data_all = _np.stack(expanded, axis=-1)
    label_all = __gather_2d__(label[:, history - 1:], generation_list)
    dates_series_all = __gather_2d__(dates_series[:, history - 1:],
                                     generation_list)

    # 去掉所有包含缺失数据的某股票某时间历史片段
    label_nan = _np.isnan(label_all)
    if _np.ndim(label_all) == 2:
        # 如果是分类问题, 需要特殊处理
        label_nan = _np.any(label_nan, axis=1)
    data_nan = _np.isnan(data_all)
    series_time_nan = _np.any(
        _np.any(data_nan, axis=2),
        axis=1
    )
    not_nan = _np.logical_not(_np.logical_or(series_time_nan, label_nan))

    data_all = data_all[not_nan]
    label_all = label_all[not_nan]
    dates_series_all = dates_series_all[not_nan]

    # max/min standardization for series greater than 1
    if normalize:
        max_val = _np.max(data_all, axis=-1, keepdims=True)
        non_rate_mask = _np.less(1.0, _np.squeeze(max_val))
        min_val = _np.min(data_all, axis=-1, keepdims=True)
        with _np.errstate(divide='ignore', invalid='ignore'):
            normalized = _np.nan_to_num((data_all - min_val)
                                        / (max_val - min_val))
        data_all[non_rate_mask] = normalized[non_rate_mask]

    data_all = _np.transpose(data_all, axes=(0, 2, 1))
    data_all = _tf.constant(data_all, dtype=_tf.float32)
    label_all = _tf.constant(label_all, dtype=_tf.float32)

    return data_all, label_all, dates_series_all


def __first_index__(array, element):
    """计算第一个出现的元素的位置."""
    return _np.min(_np.where(array == element))


@njit
def __gather_2d__(array: _np.ndarray, generation_list: _np.ndarray):
    out_shape = (len(generation_list), *array.shape[2:])
    out_array = np.empty(shape=out_shape,
                         dtype=array.dtype)
    pointer = 0
    for i, j in generation_list:
        out_array[pointer] = array[i, j]
        pointer += 1
    return out_array



from setuptools import setup

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name='alphanet',
    version='0.0.20',
    packages=['alphanet'],
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='https://github.com/Congyuwang/AlphaNetV3',
    license='MIT License',
    author='Congyu Wang',
    author_email='leonwang998@foxmail.com',
    description='A recurrent neural network for predicting '
                'stock market performance',
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    package_dir={"": "src"},
    python_requires=">=3.6",
    install_requires=[]
)
